@@ -9,7 +9,7 @@ use async_trait::async_trait;
 use destream::{de, en, EncodeSeq};
 use futures::future::{self, TryFutureExt};
 use futures::stream::{Stream, TryStreamExt};
-use log::debug;
+use log::{debug, warn};
 
 use tc_btree::Node;
 use tc_error::*;
@@ -179,6 +179,7 @@ where
     D::FileClass: From<TensorType>,
 {
     type Combine = DenseTensor<FD, FS, D, T, BlockListCombine<FD, FS, D, T, B, O>>;
+    type LeftCombine = DenseTensor<FD, FS, D, T, BlockListCombine<FD, FS, D, T, B, O>>;
 
     fn and(self, other: DenseTensor<FD, FS, D, T, O>) -> TCResult<Self::Combine> {
         self.combine(other, Array::and, Number::and, NumberType::Bool)
@@ -203,6 +204,7 @@ where
     D::FileClass: From<TensorType>,
 {
     type Combine = Tensor<FD, FS, D, T>;
+    type LeftCombine = Tensor<FD, FS, D, T>;
 
     fn and(self, other: Tensor<FD, FS, D, T>) -> TCResult<Self::Combine> {
         match other {
@@ -390,10 +392,6 @@ where
 {
     type Txn = T;
 
-    async fn mask(self, _txn: T, _other: DenseTensor<FD, FS, D, T, O>) -> TCResult<()> {
-        Err(TCError::not_implemented("DenseTensor::mask"))
-    }
-
     async fn write(
         self,
         txn: T,
@@ -417,19 +415,6 @@ where
 {
     type Txn = T;
 
-    async fn mask(self, txn: T, other: Tensor<FD, FS, D, T>) -> TCResult<()> {
-        let other = if self.shape() == other.shape() {
-            other
-        } else {
-            other.broadcast(self.shape().clone())?
-        };
-
-        match other {
-            Tensor::Dense(dense) => self.mask(txn, dense).await,
-            Tensor::Sparse(sparse) => self.mask(txn, sparse.into_dense()).await,
-        }
-    }
-
     async fn write(self, txn: T, bounds: Bounds, other: Tensor<FD, FS, D, T>) -> TCResult<()> {
         debug!("DenseTensor::write {} to {}", other, bounds);
 
@@ -459,6 +444,7 @@ where
     D::FileClass: From<TensorType>,
 {
     type Combine = DenseTensor<FD, FS, D, T, BlockListCombine<FD, FS, D, T, B, O>>;
+    type LeftCombine = DenseTensor<FD, FS, D, T, BlockListCombine<FD, FS, D, T, B, O>>;
 
     fn add(self, other: DenseTensor<FD, FS, D, T, O>) -> TCResult<Self::Combine> {
         fn add_array(l: &Array, r: &Array) -> Array {
@@ -472,6 +458,10 @@ where
 
     fn div(self, other: DenseTensor<FD, FS, D, T, O>) -> TCResult<Self::Combine> {
         fn div_array(l: &Array, r: &Array) -> Array {
+            if !r.all() {
+                warn!("divide by zero in DenseTensor::div");
+            }
+
             debug_assert_eq!(l.len(), r.len());
             l / r
         }
@@ -513,6 +503,7 @@ where
     D::FileClass: From<TensorType>,
 {
     type Combine = Tensor<FD, FS, D, T>;
+    type LeftCombine = Tensor<FD, FS, D, T>;
 
     fn add(self, other: Tensor<FD, FS, D, T>) -> TCResult<Self::Combine> {
         match other {
