@@ -32,6 +32,9 @@ mod sparse;
 mod stream;
 mod transform;
 
+const ERR_INF: &str = "Tensor combination resulted in an infinite value";
+const ERR_NAN: &str = "Tensor combination resulted in a non-numeric value";
+
 const PREFIX: PathLabel = path_label(&["state", "collection", "tensor"]);
 
 /// The file extension of a [`Tensor`]
@@ -140,8 +143,11 @@ pub trait TensorBoolean<O> {
     /// The result type of a boolean operation.
     type Combine: TensorInstance;
 
+    /// The result type of a boolean operation which may ignore right-hand values.
+    type LeftCombine: TensorInstance;
+
     /// Logical and
-    fn and(self, other: O) -> TCResult<Self::Combine>;
+    fn and(self, other: O) -> TCResult<Self::LeftCombine>;
 
     /// Logical or
     fn or(self, other: O) -> TCResult<Self::Combine>;
@@ -199,9 +205,6 @@ pub trait TensorDualIO<D: Dir, O> {
     /// The type of [`Transaction`] to expect
     type Txn: Transaction<D>;
 
-    /// Zero out the elements of this [`Tensor`] where the corresponding element of `value` is nonzero.
-    async fn mask(self, txn: Self::Txn, value: O) -> TCResult<()>;
-
     /// Overwrite the slice of this [`Tensor`] given by [`Bounds`] with the given `value`.
     async fn write(self, txn: Self::Txn, bounds: Bounds, value: O) -> TCResult<()>;
 }
@@ -211,14 +214,17 @@ pub trait TensorMath<D: Dir, O> {
     /// The result type of a math operation
     type Combine: TensorInstance;
 
+    /// The result type of a math operation which may ignore right-hand-side values
+    type LeftCombine: TensorInstance;
+
     /// Add two tensors together.
     fn add(self, other: O) -> TCResult<Self::Combine>;
 
     /// Divide `self` by `other`.
-    fn div(self, other: O) -> TCResult<Self::Combine>;
+    fn div(self, other: O) -> TCResult<Self::LeftCombine>;
 
     /// Multiply two tensors together.
-    fn mul(self, other: O) -> TCResult<Self::Combine>;
+    fn mul(self, other: O) -> TCResult<Self::LeftCombine>;
 
     /// Subtract `other` from `self`.
     fn sub(self, other: O) -> TCResult<Self::Combine>;
@@ -344,6 +350,22 @@ pub enum Tensor<FD, FS, D, T> {
     Sparse(SparseTensor<FD, FS, D, T, SparseAccessor<FD, FS, D, T>>),
 }
 
+impl<FD, FS, D, T> Tensor<FD, FS, D, T>
+where
+    FD: File<Array> + TryFrom<D::File, Error = TCError>,
+    FS: File<Node> + TryFrom<D::File, Error = TCError>,
+    D: Dir,
+    T: Transaction<D>,
+    D::FileClass: From<TensorType>,
+{
+    pub fn schema(&self) -> Schema {
+        Schema {
+            dtype: self.dtype(),
+            shape: self.shape().clone(),
+        }
+    }
+}
+
 impl<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D>> Instance for Tensor<FD, FS, D, T> {
     type Class = TensorType;
 
@@ -427,8 +449,9 @@ where
     D::FileClass: From<TensorType>,
 {
     type Combine = Self;
+    type LeftCombine = Self;
 
-    fn and(self, other: Self) -> TCResult<Self::Combine> {
+    fn and(self, other: Self) -> TCResult<Self::LeftCombine> {
         match self {
             Self::Dense(dense) => dense.and(other),
             Self::Sparse(sparse) => sparse.and(other),
@@ -550,13 +573,6 @@ where
 {
     type Txn = T;
 
-    async fn mask(self, txn: T, other: Self) -> TCResult<()> {
-        match self {
-            Self::Dense(this) => this.mask(txn, other).await,
-            Self::Sparse(this) => this.mask(txn, other).await,
-        }
-    }
-
     async fn write(self, txn: T, bounds: Bounds, value: Self) -> TCResult<()> {
         debug!("Tensor::write {} to {}", value, bounds);
 
@@ -576,6 +592,7 @@ where
     D::FileClass: From<TensorType>,
 {
     type Combine = Self;
+    type LeftCombine = Self;
 
     fn add(self, other: Self) -> TCResult<Self::Combine> {
         match self {
@@ -584,14 +601,14 @@ where
         }
     }
 
-    fn div(self, other: Self) -> TCResult<Self::Combine> {
+    fn div(self, other: Self) -> TCResult<Self::LeftCombine> {
         match self {
             Self::Dense(this) => this.div(other),
             Self::Sparse(this) => this.div(other),
         }
     }
 
-    fn mul(self, other: Self) -> TCResult<Self::Combine> {
+    fn mul(self, other: Self) -> TCResult<Self::LeftCombine> {
         match self {
             Self::Dense(this) => this.mul(other),
             Self::Sparse(this) => this.mul(other),
