@@ -1,4 +1,5 @@
 use std::convert::TryFrom;
+use std::fmt;
 
 use afarray::{Array, CoordBlocks, CoordMerge, Coords};
 use async_trait::async_trait;
@@ -11,14 +12,14 @@ use tc_btree::Node;
 use tc_error::*;
 use tc_transact::fs::{Dir, File};
 use tc_transact::{Transaction, TxnId};
-use tc_value::{Number, NumberClass, NumberInstance, NumberType};
+use tc_value::{FloatInstance, Number, NumberClass, NumberInstance, NumberType};
 use tcgeneric::{TCBoxTryFuture, TCBoxTryStream};
 
 use crate::dense::{DenseAccess, DenseAccessor, DenseTensor, PER_BLOCK};
 use crate::stream::{sorted_coords, sorted_values, Read, ReadValueAt};
-use crate::transform;
 use crate::{
-    coord_bounds, AxisBounds, Bounds, Coord, Phantom, Shape, TensorAccess, TensorType, TensorUnary,
+    coord_bounds, transform, AxisBounds, Bounds, Coord, Phantom, Shape, TensorAccess, TensorType,
+    TensorUnary, ERR_INF, ERR_NAN,
 };
 
 use super::combine::{coord_to_offset, SparseCombine};
@@ -28,7 +29,7 @@ use super::{SparseRow, SparseStream, SparseTensor};
 /// Access methods for [`SparseTensor`] data
 #[async_trait]
 pub trait SparseAccess<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D>>:
-    Clone + ReadValueAt<D, Txn = T> + TensorAccess + Send + Sync + 'static
+    Clone + fmt::Display + ReadValueAt<D, Txn = T> + TensorAccess + Send + Sync + 'static
 {
     /// The type of a slice of this accessor
     type Slice: SparseAccess<FD, FS, D, T>;
@@ -44,15 +45,15 @@ pub trait SparseAccess<FD: File<Array>, FS: File<Node>, D: Dir, T: Transaction<D
 
     /// Return an ordered stream of unique [`Coord`]s on the given axes with nonzero values.
     ///
+    /// Note: is it *not* safe to assume that all returned coordinates are filled,
+    /// but it *is* safe to assume that all coordinates *not* returned are *not* filled.
+    ///
     /// TODO: add `sort` parameter to make sorting results optional.
     async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>>;
 
     /// Return the number of nonzero values in this [`SparseTensor`].
     async fn filled_count(self, txn: T) -> TCResult<u64>;
 
-    /// Return `true` if this [`SparseTensor`] contains only zero values.
-    async fn is_empty(&self, txn: &T) -> TCResult<bool>;
-
     /// Return a slice of this accessor with the given [`Bounds`].
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice>;
 
@@ -77,6 +78,7 @@ pub enum SparseAccessor<FD, FS, D, T> {
     Broadcast(Box<SparseBroadcast<FD, FS, D, T, Self>>),
     Cast(Box<SparseCast<FD, FS, D, T, Self>>),
     Combine(Box<SparseCombinator<FD, FS, D, T, Self, Self>>),
+    CombineLeft(Box<SparseLeftCombinator<FD, FS, D, T, Self, Self>>),
     Dense(Box<DenseToSparse<FD, FS, D, T, DenseAccessor<FD, FS, D, T>>>),
     Expand(Box<SparseExpand<FD, FS, D, T, Self>>),
     Slice(SparseTableSlice<FD, FS, D, T>),
@@ -99,6 +101,7 @@ where
             Self::Broadcast(broadcast) => broadcast.dtype(),
             Self::Cast(cast) => cast.dtype(),
             Self::Combine(combine) => combine.dtype(),
+            Self::CombineLeft(combine) => combine.dtype(),
             Self::Dense(dense) => dense.dtype(),
             Self::Expand(expand) => expand.dtype(),
             Self::Slice(slice) => slice.dtype(),
@@ -114,6 +117,7 @@ where
             Self::Broadcast(broadcast) => broadcast.ndim(),
             Self::Cast(cast) => cast.ndim(),
             Self::Combine(combine) => combine.ndim(),
+            Self::CombineLeft(combine) => combine.ndim(),
             Self::Dense(dense) => dense.ndim(),
             Self::Expand(expand) => expand.ndim(),
             Self::Slice(slice) => slice.ndim(),
@@ -129,6 +133,7 @@ where
             Self::Broadcast(broadcast) => broadcast.shape(),
             Self::Cast(cast) => cast.shape(),
             Self::Combine(combine) => combine.shape(),
+            Self::CombineLeft(combine) => combine.shape(),
             Self::Dense(dense) => dense.shape(),
             Self::Expand(expand) => expand.shape(),
             Self::Reduce(reduce) => reduce.shape(),
@@ -144,6 +149,7 @@ where
             Self::Broadcast(broadcast) => broadcast.size(),
             Self::Cast(cast) => cast.size(),
             Self::Combine(combine) => combine.size(),
+            Self::CombineLeft(combine) => combine.size(),
             Self::Dense(dense) => dense.size(),
             Self::Expand(expand) => expand.size(),
             Self::Slice(slice) => slice.size(),
@@ -176,6 +182,7 @@ where
             Self::Broadcast(broadcast) => broadcast.filled(txn).await,
             Self::Cast(cast) => cast.filled(txn).await,
             Self::Combine(combine) => combine.filled(txn).await,
+            Self::CombineLeft(combine) => combine.filled(txn).await,
             Self::Dense(dense) => dense.filled(txn).await,
             Self::Expand(expand) => expand.filled(txn).await,
             Self::Reduce(reduce) => reduce.filled(txn).await,
@@ -191,6 +198,7 @@ where
             Self::Broadcast(broadcast) => broadcast.filled_at(txn, axes).await,
             Self::Cast(cast) => cast.filled_at(txn, axes).await,
             Self::Combine(combine) => combine.filled_at(txn, axes).await,
+            Self::CombineLeft(combine) => combine.filled_at(txn, axes).await,
             Self::Dense(dense) => dense.filled_at(txn, axes).await,
             Self::Expand(expand) => expand.filled_at(txn, axes).await,
             Self::Reduce(reduce) => reduce.filled_at(txn, axes).await,
@@ -206,6 +214,7 @@ where
             Self::Broadcast(broadcast) => broadcast.filled_count(txn).await,
             Self::Cast(cast) => cast.filled_count(txn).await,
             Self::Combine(combine) => combine.filled_count(txn).await,
+            Self::CombineLeft(combine) => combine.filled_count(txn).await,
             Self::Dense(dense) => dense.filled_count(txn).await,
             Self::Expand(expand) => expand.filled_count(txn).await,
             Self::Reduce(reduce) => reduce.filled_count(txn).await,
@@ -216,26 +225,12 @@ where
         }
     }
 
-    async fn is_empty(&self, txn: &T) -> TCResult<bool> {
-        match self {
-            Self::Broadcast(broadcast) => broadcast.is_empty(txn).await,
-            Self::Cast(cast) => cast.is_empty(txn).await,
-            Self::Combine(combine) => combine.is_empty(txn).await,
-            Self::Dense(dense) => dense.is_empty(txn).await,
-            Self::Expand(expand) => expand.is_empty(txn).await,
-            Self::Reduce(reduce) => reduce.is_empty(txn).await,
-            Self::Slice(slice) => slice.is_empty(txn).await,
-            Self::Table(table) => table.is_empty(txn).await,
-            Self::Transpose(transpose) => transpose.is_empty(txn).await,
-            Self::Unary(unary) => unary.is_empty(txn).await,
-        }
-    }
-
     fn slice(self, bounds: Bounds) -> TCResult<Self> {
         match self {
             Self::Broadcast(broadcast) => broadcast.slice(bounds).map(SparseAccess::accessor),
             Self::Cast(cast) => cast.slice(bounds).map(SparseAccess::accessor),
             Self::Combine(combinator) => combinator.slice(bounds).map(SparseAccess::accessor),
+            Self::CombineLeft(combinator) => combinator.slice(bounds).map(SparseAccess::accessor),
             Self::Dense(dense) => dense.slice(bounds).map(SparseAccess::accessor),
             Self::Expand(expand) => expand.slice(bounds).map(SparseAccess::accessor),
             Self::Reduce(reduce) => reduce.slice(bounds).map(SparseAccess::accessor),
@@ -255,6 +250,9 @@ where
             Self::Combine(combinator) => combinator
                 .transpose(permutation)
                 .map(SparseAccess::accessor),
+            Self::CombineLeft(combinator) => combinator
+                .transpose(permutation)
+                .map(SparseAccess::accessor),
             Self::Dense(dense) => dense.transpose(permutation).map(SparseAccess::accessor),
             Self::Expand(expand) => expand.transpose(permutation).map(SparseAccess::accessor),
             Self::Reduce(reduce) => reduce.transpose(permutation).map(SparseAccess::accessor),
@@ -300,6 +298,7 @@ where
             Self::Broadcast(broadcast) => broadcast.read_value_at(txn, coord),
             Self::Cast(cast) => cast.read_value_at(txn, coord),
             Self::Combine(combine) => combine.read_value_at(txn, coord),
+            Self::CombineLeft(combine) => combine.read_value_at(txn, coord),
             Self::Dense(dense) => dense.read_value_at(txn, coord),
             Self::Expand(expand) => expand.read_value_at(txn, coord),
             Self::Reduce(reduce) => reduce.read_value_at(txn, coord),
@@ -311,6 +310,24 @@ where
     }
 }
 
+impl<FD, FS, D, T> fmt::Display for SparseAccessor<FD, FS, D, T> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        match self {
+            Self::Broadcast(broadcast) => fmt::Display::fmt(broadcast, f),
+            Self::Cast(cast) => fmt::Display::fmt(cast, f),
+            Self::Combine(combinator) => fmt::Display::fmt(combinator, f),
+            Self::CombineLeft(combinator) => fmt::Display::fmt(combinator, f),
+            Self::Dense(dense) => fmt::Display::fmt(dense, f),
+            Self::Expand(expand) => fmt::Display::fmt(expand, f),
+            Self::Reduce(reduce) => fmt::Display::fmt(reduce, f),
+            Self::Slice(slice) => fmt::Display::fmt(slice, f),
+            Self::Table(table) => fmt::Display::fmt(table, f),
+            Self::Transpose(transpose) => fmt::Display::fmt(transpose, f),
+            Self::Unary(unary) => fmt::Display::fmt(unary, f),
+        }
+    }
+}
+
 #[derive(Clone)]
 pub struct DenseToSparse<FD, FS, D, T, B> {
     source: B,
@@ -430,17 +447,6 @@ where
             .await
     }
 
-    async fn is_empty(&self, txn: &T) -> TCResult<bool> {
-        let mut blocks = self.source.clone().block_stream(txn.clone()).await?;
-        while let Some(block) = blocks.try_next().await? {
-            if block.any() {
-                return Ok(false);
-            }
-        }
-
-        Ok(true)
-    }
-
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         self.source.slice(bounds).map(DenseToSparse::from)
     }
@@ -465,6 +471,12 @@ where
     }
 }
 
+impl<FD, FS, D, T, B> fmt::Display for DenseToSparse<FD, FS, D, T, B> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("a sparse representation of a dense Tensor")
+    }
+}
+
 impl<FD, FS, D, T, B> From<B> for DenseToSparse<FD, FS, D, T, B> {
     fn from(source: B) -> Self {
         Self {
@@ -550,11 +562,6 @@ where
     async fn filled<'a>(self, txn: T) -> TCResult<SparseStream<'a>> {
         debug!("SparseBroadcast::filled");
 
-        if self.is_empty(&txn).await? {
-            debug!("sparse tensor to broadcast is empty");
-            return Ok(Box::pin(stream::empty()));
-        }
-
         let source_axes: Vec<usize> = (0..self.source.ndim()).collect();
         let ndim = self.ndim();
 
@@ -569,7 +576,10 @@ where
         let coords = filled_at
             .map_ok(move |coords| stream::iter(coords.to_vec()).map(TCResult::Ok))
             .try_flatten()
-            .map_ok(move |coord| rebase.map_coord(coord))
+            .map_ok(move |coord| {
+                debug!("SparseBroadcast::filled at source coord {:?}", coord);
+                rebase.map_coord(coord)
+            })
             .map_ok(move |bounds| stream::iter(bounds.affected().map(TCResult::Ok)))
             .try_flatten();
 
@@ -579,13 +589,17 @@ where
     }
 
     async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
-        debug!("SparseBroadcast::filled_at {:?}", axes);
-        self.shape().validate_axes(&axes)?;
+        debug!(
+            "SparseBroadcast::filled_at {:?} (source is {})",
+            axes, self.source
+        );
 
-        if axes.is_empty() || self.is_empty(&txn).await? {
+        if axes.is_empty() {
             return Ok(Box::pin(stream::empty()));
         }
 
+        self.shape().validate_axes(&axes)?;
+
         let shape = Shape::from({
             let shape = self.shape();
             axes.iter().map(|x| shape[*x]).collect::<Vec<u64>>()
@@ -599,7 +613,10 @@ where
         let filled_at = filled_at
             .map_ok(|coords| stream::iter(coords.to_vec()).map(TCResult::Ok))
             .try_flatten()
-            .map_ok(move |coord| rebase.map_coord(coord))
+            .map_ok(move |coord| {
+                debug!("broadcast source coord {:?}", coord);
+                rebase.map_coord(coord)
+            })
             // TODO: can this happen in `Coords`? maybe a new stream generator type?
             .map_ok(move |bounds| stream::iter(bounds.affected().map(TCResult::Ok)))
             .try_flatten();
@@ -623,16 +640,19 @@ where
             .await
     }
 
-    async fn is_empty(&self, txn: &T) -> TCResult<bool> {
-        debug!("SparseBroadcast::is_empty?");
-        self.source.is_empty(txn).await
-    }
-
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
+        debug!("SparseBroadcast::slice {} from {}", bounds, self.shape());
         self.shape().validate_bounds(&bounds)?;
 
         let shape = bounds.to_shape(self.shape())?;
         let source_bounds = self.rebase.invert_bounds(bounds);
+
+        debug!(
+            "SparseBroadcast::slice source bounds are {} (from {})",
+            source_bounds,
+            self.source.shape()
+        );
+
         let source = self.source.slice(source_bounds)?;
         SparseBroadcast::new(source, shape)
     }
@@ -666,6 +686,12 @@ where
     }
 }
 
+impl<FD, FS, D, T, A> fmt::Display for SparseBroadcast<FD, FS, D, T, A> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("a broadcasted sparse Tensor")
+    }
+}
+
 #[derive(Clone)]
 pub struct SparseCast<FD, FS, D, T, A> {
     source: A,
@@ -743,10 +769,6 @@ where
         self.source.filled_count(txn).await
     }
 
-    async fn is_empty(&self, txn: &T) -> TCResult<bool> {
-        self.source.is_empty(txn).await
-    }
-
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         Ok(SparseCast {
             source: self.source.slice(bounds)?,
@@ -786,12 +808,17 @@ where
     }
 }
 
+impl<FD, FS, D, T, A> fmt::Display for SparseCast<FD, FS, D, T, A> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("a casted sparse Tensor")
+    }
+}
+
 #[derive(Clone)]
 pub struct SparseCombinator<FD, FS, D, T, L, R> {
     left: L,
     right: R,
     combinator: fn(Number, Number) -> Number,
-    dtype: NumberType,
     phantom: Phantom<FD, FS, D, T>,
 }
 
@@ -804,12 +831,7 @@ where
     L: SparseAccess<FD, FS, D, T>,
     R: SparseAccess<FD, FS, D, T>,
 {
-    pub fn new(
-        left: L,
-        right: R,
-        combinator: fn(Number, Number) -> Number,
-        dtype: NumberType,
-    ) -> TCResult<Self> {
+    pub fn new(left: L, right: R, combinator: fn(Number, Number) -> Number) -> TCResult<Self> {
         if left.shape() != right.shape() {
             return Err(TCError::unsupported(
                 "tried to combine SparseTensors with different shapes",
@@ -820,7 +842,6 @@ where
             left,
             right,
             combinator,
-            dtype,
             phantom: Phantom::default(),
         })
     }
@@ -836,17 +857,29 @@ where
         let right_zero = self.right.dtype().zero();
 
         let offset = move |row: &SparseRow| coord_to_offset(&row.0, &coord_bounds);
-        let combined = SparseCombine::new(left, right, offset).map_ok(move |(l, r)| match (l, r) {
-            (Some((l_coord, l)), Some((r_coord, r))) => {
-                debug_assert_eq!(l_coord, r_coord);
-                (l_coord, combinator(l, r))
-            }
-            (Some((l_coord, l)), None) => (l_coord, combinator(l, right_zero)),
-            (None, Some((r_coord, r))) => (r_coord, combinator(left_zero, r)),
-            (None, None) => {
-                panic!("expected a coordinate and value from one sparse tensor stream")
-            }
-        });
+        let combined = SparseCombine::new(left, right, offset)
+            .map_ok(move |(l, r)| match (l, r) {
+                (Some((l_coord, l)), Some((r_coord, r))) => {
+                    debug_assert_eq!(l_coord, r_coord);
+                    (l_coord, combinator(l, r))
+                }
+                (Some((l_coord, l)), None) => (l_coord, combinator(l, right_zero)),
+                (None, Some((r_coord, r))) => (r_coord, combinator(left_zero, r)),
+                (None, None) => {
+                    panic!("expected a coordinate and value from one sparse tensor stream")
+                }
+            })
+            .map(|result| {
+                result.and_then(|(coord, n)| {
+                    if n.is_infinite() {
+                        Err(TCError::unsupported(ERR_INF))
+                    } else if n.is_nan() {
+                        Err(TCError::unsupported(ERR_NAN))
+                    } else {
+                        Ok((coord, n))
+                    }
+                })
+            });
 
         Ok(Box::pin(combined))
     }
@@ -862,7 +895,7 @@ where
     R: SparseAccess<FD, FS, D, T>,
 {
     fn dtype(&self) -> NumberType {
-        self.dtype
+        (self.combinator)(self.left.dtype().zero(), self.right.dtype().zero()).class()
     }
 
     fn ndim(&self) -> usize {
@@ -896,7 +929,6 @@ where
             left: self.left.accessor(),
             right: self.right.accessor(),
             combinator: self.combinator,
-            dtype: self.dtype,
             phantom: Phantom::default(),
         }))
     }
@@ -904,7 +936,8 @@ where
     async fn filled<'a>(self, txn: T) -> TCResult<SparseStream<'a>> {
         let zero = self.dtype().zero();
         let filled_inner = self.filled_inner(txn).await?;
-        let filled = filled_inner.try_filter(move |(_, value)| future::ready(value != &zero));
+        let filled = filled_inner.try_filter(move |(_, value)| future::ready(*value != zero));
+
         Ok(Box::pin(filled))
     }
 
@@ -937,11 +970,6 @@ where
             .await
     }
 
-    async fn is_empty(&self, txn: &T) -> TCResult<bool> {
-        let mut filled = self.clone().filled(txn.clone()).await?;
-        filled.try_next().map_ok(|v| v.is_some()).await
-    }
-
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         debug!("SparseCombinator::slice {}", bounds);
 
@@ -953,7 +981,6 @@ where
             left,
             right,
             combinator: self.combinator,
-            dtype: self.dtype,
             phantom: self.phantom,
         })
     }
@@ -967,7 +994,6 @@ where
             left,
             right,
             combinator: self.combinator,
-            dtype: self.dtype,
             phantom: self.phantom,
         })
     }
@@ -995,6 +1021,205 @@ where
     }
 }
 
+impl<FD, FS, D, T, L, R> fmt::Display for SparseCombinator<FD, FS, D, T, L, R> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("the result of combining two sparse Tensors")
+    }
+}
+
+#[derive(Clone)]
+pub struct SparseLeftCombinator<FD, FS, D, T, L, R> {
+    left: L,
+    right: R,
+    combinator: fn(Number, Number) -> Number,
+    phantom: Phantom<FD, FS, D, T>,
+}
+
+impl<FD, FS, D, T, L, R> SparseLeftCombinator<FD, FS, D, T, L, R>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    L: SparseAccess<FD, FS, D, T>,
+    R: SparseAccess<FD, FS, D, T>,
+{
+    pub fn new(left: L, right: R, combinator: fn(Number, Number) -> Number) -> TCResult<Self> {
+        if left.shape() != right.shape() {
+            return Err(TCError::unsupported(
+                "tried to combine SparseTensors with different shapes",
+            ));
+        }
+
+        Ok(SparseLeftCombinator {
+            left,
+            right,
+            combinator,
+            phantom: Phantom::default(),
+        })
+    }
+
+    pub async fn filled_inner<'a>(self, txn: T) -> TCResult<SparseStream<'a>> {
+        debug!(
+            "SparseLeftCombinator::filled_inner ({}, {})",
+            self.left, self.right
+        );
+
+        let left = self.left.filled(txn.clone()).await?;
+
+        let combinator = self.combinator;
+        let right = self.right;
+
+        let filled = left
+            .and_then(move |(coord, left_value)| {
+                right
+                    .clone()
+                    .read_value_at(txn.clone(), coord)
+                    .map_ok(move |(coord, right_value)| (coord, left_value, right_value))
+            })
+            .map_ok(move |(coord, left, right)| (coord, combinator(left, right)));
+
+        Ok(Box::pin(filled))
+    }
+}
+
+impl<FD, FS, D, T, L, R> TensorAccess for SparseLeftCombinator<FD, FS, D, T, L, R>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    L: SparseAccess<FD, FS, D, T>,
+    R: SparseAccess<FD, FS, D, T>,
+{
+    fn dtype(&self) -> NumberType {
+        (self.combinator)(self.left.dtype().zero(), self.right.dtype().zero()).class()
+    }
+
+    fn ndim(&self) -> usize {
+        self.left.ndim()
+    }
+
+    fn shape(&'_ self) -> &'_ Shape {
+        self.left.shape()
+    }
+
+    fn size(&self) -> u64 {
+        self.left.size()
+    }
+}
+
+#[async_trait]
+impl<FD, FS, D, T, L, R> SparseAccess<FD, FS, D, T> for SparseLeftCombinator<FD, FS, D, T, L, R>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    L: SparseAccess<FD, FS, D, T>,
+    R: SparseAccess<FD, FS, D, T>,
+{
+    type Slice = SparseLeftCombinator<FD, FS, D, T, L::Slice, R::Slice>;
+    type Transpose = SparseLeftCombinator<FD, FS, D, T, L::Transpose, R::Transpose>;
+
+    fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
+        SparseAccessor::CombineLeft(Box::new(SparseLeftCombinator {
+            left: self.left.accessor(),
+            right: self.right.accessor(),
+            combinator: self.combinator,
+            phantom: Phantom::default(),
+        }))
+    }
+
+    async fn filled<'a>(self, txn: T) -> TCResult<SparseStream<'a>> {
+        let zero = self.dtype().zero();
+        let filled_inner = self.filled_inner(txn).await?;
+        let filled = filled_inner.try_filter(move |(_, value)| future::ready(value != &zero));
+        Ok(Box::pin(filled))
+    }
+
+    async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        self.left.filled_at(txn, axes).await
+    }
+
+    async fn filled_count(self, txn: T) -> TCResult<u64> {
+        let filled = self.filled(txn).await?;
+
+        filled
+            .try_fold(0u64, |count, _| future::ready(Ok(count + 1)))
+            .await
+    }
+
+    fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
+        debug!("SparseLeftCombinator::slice left {} {}", self.left, bounds);
+        let left = self.left.slice(bounds.clone())?;
+
+        debug!(
+            "SparseLeftCombinator::slice right {} {}",
+            self.right, bounds
+        );
+        let right = self.right.slice(bounds)?;
+
+        assert_eq!(left.shape(), right.shape());
+
+        Ok(SparseLeftCombinator {
+            left,
+            right,
+            combinator: self.combinator,
+            phantom: self.phantom,
+        })
+    }
+
+    fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
+        let left = self.left.transpose(permutation.clone())?;
+        let right = self.right.transpose(permutation)?;
+        assert_eq!(left.shape(), right.shape());
+
+        Ok(SparseLeftCombinator {
+            left,
+            right,
+            combinator: self.combinator,
+            phantom: self.phantom,
+        })
+    }
+}
+
+impl<FD, FS, D, T, L, R> ReadValueAt<D> for SparseLeftCombinator<FD, FS, D, T, L, R>
+where
+    FD: File<Array>,
+    FS: File<Node>,
+    D: Dir,
+    T: Transaction<D>,
+    L: SparseAccess<FD, FS, D, T>,
+    R: SparseAccess<FD, FS, D, T>,
+{
+    type Txn = T;
+
+    fn read_value_at<'a>(self, txn: T, coord: Coord) -> Read<'a> {
+        Box::pin(async move {
+            let left_zero = self.left.dtype().zero();
+            let zero = self.dtype().zero();
+
+            let (coord, left) = self.left.read_value_at(txn.clone(), coord).await?;
+            let (coord, value) = if left == left_zero {
+                (coord, zero)
+            } else {
+                let (coord, right) = self.right.read_value_at(txn, coord).await?;
+                let value = (self.combinator)(left, right);
+                (coord, value)
+            };
+
+            Ok((coord, value))
+        })
+    }
+}
+
+impl<FD, FS, D, T, L, R> fmt::Display for SparseLeftCombinator<FD, FS, D, T, L, R> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("the result of combining two sparse Tensors, ignoring the right Tensor where the left Tensor is zero")
+    }
+}
+
 #[derive(Clone)]
 pub struct SparseExpand<FD, FS, D, T, A> {
     source: A,
@@ -1048,13 +1273,14 @@ where
 #[async_trait]
 impl<FD, FS, D, T, A> SparseAccess<FD, FS, D, T> for SparseExpand<FD, FS, D, T, A>
 where
-    FD: File<Array>,
-    FS: File<Node>,
+    FD: File<Array> + TryFrom<D::File, Error = TCError>,
+    FS: File<Node> + TryFrom<D::File, Error = TCError>,
     D: Dir,
     T: Transaction<D>,
     A: SparseAccess<FD, FS, D, T>,
+    D::FileClass: From<TensorType>,
 {
-    type Slice = A::Slice;
+    type Slice = SparseAccessor<FD, FS, D, T>;
     type Transpose = SparseExpand<FD, FS, D, T, A::Transpose>;
 
     fn accessor(self) -> SparseAccessor<FD, FS, D, T> {
@@ -1073,6 +1299,8 @@ where
     }
 
     async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        debug!("SparseExpand::filled_at {:?}", axes);
+
         self.shape().validate_axes(&axes)?;
 
         if axes.is_empty() {
@@ -1090,6 +1318,11 @@ where
             }
         };
 
+        debug!(
+            "SparseExpand::filled_at {:?} will expand axis {:?}",
+            axes, expand
+        );
+
         let source_axes = self.rebase.invert_axes(axes);
         let source = self.source.filled_at(txn, source_axes).await?;
         if let Some(x) = expand {
@@ -1104,25 +1337,50 @@ where
         self.source.filled_count(txn).await
     }
 
-    async fn is_empty(&self, txn: &T) -> TCResult<bool> {
-        self.source.is_empty(txn).await
-    }
-
     fn slice(self, mut bounds: Bounds) -> TCResult<Self::Slice> {
         debug!("SparseExpand {} slice {}", self.shape(), bounds);
         self.shape().validate_bounds(&bounds)?;
         bounds.normalize(self.shape());
 
-        let source_bounds = self.rebase.invert_bounds(bounds);
+        if bounds == Bounds::all(self.shape()) {
+            return Ok(self.accessor());
+        }
+
+        let source_bounds = self.rebase.invert_bounds(bounds.clone());
+        let source_slice_shape = source_bounds.to_shape(self.source.shape())?;
+
+        let expand_axis = self.rebase.expand_axis();
+        let source_expand_axis = if bounds[expand_axis].is_index() {
+            // in this case the expanded dimension is elided
+            None
+        } else if source_slice_shape.len() == 1 {
+            // in this case only the expanded dimension is present in the slice
+            None
+        } else {
+            let num_elided = bounds[..expand_axis]
+                .iter()
+                .filter(|bound| bound.is_index())
+                .count();
+
+            Some(expand_axis - num_elided)
+        };
 
         debug!(
-            "slice source {} with bounds {} (removed axis {})",
+            "SparseExpand slice source {} with bounds {} (removed axis {}, expanding axis {:?})",
             self.source.shape(),
             source_bounds,
-            self.rebase.expand_axis()
+            expand_axis,
+            source_expand_axis,
         );
 
-        self.source.slice(source_bounds)
+        let slice = self.source.slice(source_bounds)?;
+        debug_assert_eq!(&source_slice_shape, slice.shape());
+
+        if let Some(source_expand_axis) = source_expand_axis {
+            SparseExpand::new(slice, source_expand_axis).map(|expansion| expansion.accessor())
+        } else {
+            Ok(slice.accessor())
+        }
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
@@ -1166,6 +1424,12 @@ where
     }
 }
 
+impl<FD, FS, D, T, A> fmt::Display for SparseExpand<FD, FS, D, T, A> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("a sparse Tensor expansion")
+    }
+}
+
 type Reductor<FD, FS, D, T> =
     fn(&SparseTensor<FD, FS, D, T, SparseAccessor<FD, FS, D, T>>, T) -> TCBoxTryFuture<Number>;
 
@@ -1239,6 +1503,8 @@ where
     }
 
     async fn filled<'a>(self, txn: T) -> TCResult<SparseStream<'a>> {
+        debug!("SparseReduce::filled");
+
         let filled_at = self
             .clone()
             .filled_at(txn.clone(), (0..self.ndim()).collect())
@@ -1270,6 +1536,8 @@ where
     }
 
     async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        debug!("SparseReduce::filled_at {:?}", axes);
+
         let source_axes = self.rebase.invert_axes(axes);
         self.source.filled_at(txn, source_axes).await
     }
@@ -1283,11 +1551,6 @@ where
             .await
     }
 
-    async fn is_empty(&self, txn: &T) -> TCResult<bool> {
-        let mut filled = self.clone().filled(txn.clone()).await?;
-        filled.try_next().map_ok(|v| v.is_some()).await
-    }
-
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         debug!("SparseReduce::slice {}", bounds);
 
@@ -1295,9 +1558,20 @@ where
 
         let reduce_axis = self.rebase.reduce_axis(&bounds);
         let source_bounds = self.rebase.invert_bounds(bounds);
-        debug!("source bounds are {}", source_bounds);
+        debug!(
+            "SparseReduce::slice source is {}, bounds are {}, source axis to reduce is {}",
+            self.source, source_bounds, reduce_axis
+        );
+
         let source = self.source.slice(source_bounds)?;
-        SparseReduce::new(source.into(), reduce_axis, self.reductor)
+        let reduced = SparseReduce::new(source.into(), reduce_axis, self.reductor)?;
+        if reduced.ndim() == 0 {
+            Err(TCError::unsupported(
+                "cannot return a zero-dimensional slice from a reduced Tensor",
+            ))
+        } else {
+            Ok(reduced)
+        }
     }
 
     fn transpose(self, permutation: Option<Vec<usize>>) -> TCResult<Self::Transpose> {
@@ -1326,6 +1600,12 @@ where
     }
 }
 
+impl<FD, FS, D, T> fmt::Display for SparseReduce<FD, FS, D, T> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("a sparse Tensor reduction")
+    }
+}
+
 #[derive(Clone)]
 pub struct SparseTranspose<FD, FS, D, T, A> {
     source: A,
@@ -1398,10 +1678,6 @@ where
     }
 
     async fn filled<'a>(self, txn: T) -> TCResult<SparseStream<'a>> {
-        if self.is_empty(&txn).await? {
-            return Ok(Box::pin(stream::empty()));
-        }
-
         let rebase = self.rebase.clone();
         let source_axes = self.rebase.invert_axes((0..self.ndim()).collect());
         let filled_at = self
@@ -1416,11 +1692,19 @@ where
     }
 
     async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        debug!(
+            "SparseTranspose::filled_at {:?} (source is {}, shape is {})",
+            axes,
+            self.source,
+            self.shape()
+        );
+
         if axes.is_empty() {
             return Ok(Box::pin(stream::empty()));
         }
 
-        let shape = self.shape().clone();
+        let shape = self.shape();
+        let shape = axes.iter().map(|x| shape[*x]).collect();
         let source_axes = self.rebase.invert_axes(axes);
         let permutation = self.rebase.map_axes(&source_axes);
         let source = self.source.filled_at(txn.clone(), source_axes).await?;
@@ -1433,10 +1717,6 @@ where
         self.source.filled_count(txn).await
     }
 
-    async fn is_empty(&self, txn: &T) -> TCResult<bool> {
-        self.source.is_empty(txn).await
-    }
-
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         debug!("SparseTranspose::slice {}", bounds);
         self.shape().validate_bounds(&bounds)?;
@@ -1481,6 +1761,12 @@ where
     }
 }
 
+impl<FD, FS, D, T, A> fmt::Display for SparseTranspose<FD, FS, D, T, A> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("a transposed sparse Tensor")
+    }
+}
+
 #[derive(Clone)]
 pub struct SparseUnary<FD, FS, D, T> {
     source: SparseAccessor<FD, FS, D, T>, // TODO: can this be a type parameter A: SparseAccess?
@@ -1558,10 +1844,6 @@ where
         self.source.filled_count(txn).await
     }
 
-    async fn is_empty(&self, txn: &T) -> TCResult<bool> {
-        self.source.is_empty(txn).await
-    }
-
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
         let source = self.source.slice(bounds)?;
         Ok(SparseUnary {
@@ -1601,3 +1883,9 @@ where
         Box::pin(read)
     }
 }
+
+impl<FD, FS, D, T> fmt::Display for SparseUnary<FD, FS, D, T> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("a unary operation on a sparse Tensor")
+    }
+}
