@@ -69,7 +69,6 @@ where
         self,
         other: SparseTensor<FD, FS, D, T, R>,
         combinator: fn(Number, Number) -> Number,
-        dtype: NumberType,
     ) -> TCResult<SparseTensor<FD, FS, D, T, SparseCombinator<FD, FS, D, T, A, R>>> {
         if self.shape() != other.shape() {
             return Err(TCError::unsupported(format!(
@@ -79,7 +78,7 @@ where
             )));
         }
 
-        let accessor = SparseCombinator::new(self.accessor, other.accessor, combinator, dtype)?;
+        let accessor = SparseCombinator::new(self.accessor, other.accessor, combinator)?;
 
         Ok(SparseTensor {
             accessor,
@@ -103,12 +102,35 @@ where
             )));
         }
 
-        let accessor =
-            SparseCombinator::new(self.accessor, other.accessor, condensor, NumberType::Bool)?;
+        let accessor = SparseCombinator::new(self.accessor, other.accessor, condensor)?;
 
         let dense = BlockListSparse::from(accessor);
         Ok(dense.into())
     }
+
+    fn left_combine<R>(
+        self,
+        other: SparseTensor<FD, FS, D, T, R>,
+        combinator: fn(Number, Number) -> Number,
+    ) -> TCResult<SparseTensor<FD, FS, D, T, SparseLeftCombinator<FD, FS, D, T, A, R>>>
+    where
+        R: SparseAccess<FD, FS, D, T>,
+    {
+        if self.shape() != other.shape() {
+            return Err(TCError::unsupported(format!(
+                "cannot compare Tensors of different shapes: {}, {}",
+                self.shape(),
+                other.shape()
+            )));
+        }
+
+        let accessor = SparseLeftCombinator::new(self.accessor, other.accessor, combinator)?;
+
+        Ok(SparseTensor {
+            accessor,
+            phantom: self.phantom,
+        })
+    }
 }
 
 impl<FD, FS, D, T> SparseTensor<FD, FS, D, T, SparseTable<FD, FS, D, T>>
@@ -177,17 +199,18 @@ where
     R: SparseAccess<FD, FS, D, T>,
 {
     type Combine = SparseTensor<FD, FS, D, T, SparseCombinator<FD, FS, D, T, L, R>>;
+    type LeftCombine = SparseTensor<FD, FS, D, T, SparseLeftCombinator<FD, FS, D, T, L, R>>;
 
-    fn and(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::Combine> {
-        self.combine(other, Number::and, NumberType::Bool)
+    fn and(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::LeftCombine> {
+        self.left_combine(other, Number::and)
     }
 
     fn or(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::Combine> {
-        self.combine(other, Number::or, NumberType::Bool)
+        self.combine(other, Number::or)
     }
 
     fn xor(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::Combine> {
-        self.combine(other, Number::xor, NumberType::Bool)
+        self.combine(other, Number::xor)
     }
 }
 
@@ -202,6 +225,7 @@ where
     D::FileClass: From<TensorType>,
 {
     type Combine = Tensor<FD, FS, D, T>;
+    type LeftCombine = Tensor<FD, FS, D, T>;
 
     fn and(self, other: Tensor<FD, FS, D, T>) -> TCResult<Self::Combine> {
         match other {
@@ -252,7 +276,7 @@ where
             (l > r).into()
         }
 
-        self.combine(other, gt, NumberType::Bool)
+        self.combine(other, gt)
     }
 
     fn gte(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::Dense> {
@@ -268,7 +292,7 @@ where
             (l < r).into()
         }
 
-        self.combine(other, lt, NumberType::Bool)
+        self.combine(other, lt)
     }
 
     fn lte(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::Dense> {
@@ -284,7 +308,7 @@ where
             (l != r).into()
         }
 
-        self.combine(other, ne, NumberType::Bool)
+        self.combine(other, ne)
     }
 }
 
@@ -357,27 +381,6 @@ where
 {
     type Txn = T;
 
-    async fn mask(self, txn: T, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<()> {
-        if self.shape() != other.shape() {
-            return Err(TCError::unsupported(format!(
-                "cannot use a Tensor with shape {} as a mask for a Tensor with shape {}",
-                other.shape(),
-                self.shape(),
-            )));
-        }
-
-        let zero = self.dtype().zero();
-        let txn_id = *txn.id();
-
-        let filled = other.accessor.filled(txn).await?;
-
-        filled
-            .map_ok(|(coord, _)| self.write_value_at(txn_id, coord, zero.clone()))
-            .try_buffer_unordered(num_cpus::get())
-            .try_fold((), |_, _| future::ready(Ok(())))
-            .await
-    }
-
     async fn write(
         self,
         txn: T,
@@ -421,19 +424,6 @@ where
 {
     type Txn = T;
 
-    async fn mask(self, txn: Self::Txn, other: Tensor<FD, FS, D, T>) -> TCResult<()> {
-        let other = if self.shape() == other.shape() {
-            other
-        } else {
-            other.broadcast(self.shape().clone())?
-        };
-
-        match other {
-            Tensor::Dense(other) => self.mask(txn, other.into_sparse()).await,
-            Tensor::Sparse(other) => self.mask(txn, other).await,
-        }
-    }
-
     async fn write(
         self,
         txn: Self::Txn,
@@ -502,27 +492,26 @@ where
     R: SparseAccess<FD, FS, D, T>,
 {
     type Combine = SparseTensor<FD, FS, D, T, SparseCombinator<FD, FS, D, T, L, R>>;
+    type LeftCombine = SparseTensor<FD, FS, D, T, SparseLeftCombinator<FD, FS, D, T, L, R>>;
 
     fn add(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::Combine> {
-        let dtype = Ord::max(self.dtype(), other.dtype());
-        self.combine(other, Number::add, dtype)
+        debug!("SparseTensor::add");
+        self.combine(other, Number::add)
     }
 
-    fn div(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::Combine> {
-        // TODO: implement left_combine
-        let dtype = Ord::max(self.dtype(), other.dtype());
-        self.combine(other, Number::div, dtype)
+    fn div(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::LeftCombine> {
+        debug!("SparseTensor::div");
+        self.left_combine(other, Number::div)
     }
 
-    fn mul(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::Combine> {
-        // TODO: implement left_combine
-        let dtype = Ord::max(self.dtype(), other.dtype());
-        self.combine(other, Number::mul, dtype)
+    fn mul(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::LeftCombine> {
+        debug!("SparseTensor::mul");
+        self.left_combine(other, Number::mul)
     }
 
     fn sub(self, other: SparseTensor<FD, FS, D, T, R>) -> TCResult<Self::Combine> {
-        let dtype = Ord::max(self.dtype(), other.dtype());
-        self.combine(other, Number::sub, dtype)
+        debug!("SparseTensor::sub");
+        self.combine(other, Number::sub)
     }
 }
 
@@ -536,6 +525,7 @@ where
     D::FileClass: From<TensorType>,
 {
     type Combine = Tensor<FD, FS, D, T>;
+    type LeftCombine = Tensor<FD, FS, D, T>;
 
     fn add(self, other: Tensor<FD, FS, D, T>) -> TCResult<Self::Combine> {
         match other {
