@@ -1,5 +1,6 @@
 use std::collections::HashMap;
 use std::convert::{TryFrom, TryInto};
+use std::fmt;
 use std::marker::PhantomData;
 
 use afarray::{Array, CoordBlocks, Coords};
@@ -114,13 +115,15 @@ where
     }
 
     async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
+        debug!("SparseTable::filled_at {:?}", axes);
+
         self.shape().validate_axes(&axes)?;
 
-        if axes.is_empty() || self.is_empty(&txn).await? {
+        if axes.is_empty() {
             return Ok(Box::pin(stream::empty()));
         }
 
-        let sort = axes.iter().zip(0..axes.len()).any(|(x, y)| x != &y);
+        let sort = axes.iter().enumerate().any(|(x, y)| &x != y);
 
         let shape = self.shape();
         let shape = axes.iter().map(|x| shape[*x]).collect::<Vec<u64>>();
@@ -129,9 +132,11 @@ where
         let coords = CoordBlocks::new(coords, shape.len(), PER_BLOCK);
 
         if sort {
+            debug!("SparseTable::filled_at shape is {:?}", shape);
             let coords = sorted_coords::<FD, FS, D, T, _>(&txn, shape.into(), coords).await?;
             Ok(Box::pin(coords))
         } else {
+            debug!("SparseTable::filled_at has no need to sort");
             Ok(Box::pin(coords))
         }
     }
@@ -140,14 +145,17 @@ where
         self.table.count(*txn.id()).await
     }
 
-    async fn is_empty(&self, txn: &T) -> TCResult<bool> {
-        self.table.is_empty(txn).await
-    }
-
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
-        debug!("SparseTable {} slice {}", self.shape(), bounds);
         self.shape().validate_bounds(&bounds)?;
         let table_bounds = table_bounds(self.shape(), &bounds)?;
+
+        debug!(
+            "SparseTable {} slice {}, table bounds are {}",
+            self.shape(),
+            bounds,
+            table_bounds
+        );
+
         if table_bounds.is_empty() {
             Ok(self.accessor())
         } else {
@@ -196,6 +204,12 @@ where
     }
 }
 
+impl<FD, FS, D, T> fmt::Display for SparseTable<FD, FS, D, T> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("a sparse Tensor's underlying Table representation")
+    }
+}
+
 #[async_trait]
 impl<FD, FS, D, T> Transact for SparseTable<FD, FS, D, T>
 where
@@ -229,6 +243,8 @@ where
         store: D,
         txn: &T,
     ) -> TCResult<Self> {
+        debug!("SparseTable::copy_from {}", instance.accessor);
+
         let txn_id = *txn.id();
         let shape = instance.shape().clone();
         let dtype = instance.dtype();
@@ -325,6 +341,12 @@ impl<FD, FS, D, T> SparseTableSlice<FD, FS, D, T> {
         table: Merged<FS, D, T>,
         rebase: transform::Slice,
     ) -> Self {
+        debug!(
+            "SparseTableSlice::new {} from {}",
+            rebase.bounds(),
+            source.shape()
+        );
+
         Self {
             source,
             table,
@@ -368,6 +390,8 @@ where
     }
 
     async fn filled<'a>(self, txn: T) -> TCResult<SparseStream<'a>> {
+        debug!("SparseTableSlice::filled");
+
         let rebase = self.rebase;
         let rows = self.table.rows(*txn.id()).await?;
         let filled = rows
@@ -379,10 +403,9 @@ where
     }
 
     async fn filled_at<'a>(self, txn: T, axes: Vec<usize>) -> TCResult<TCBoxTryStream<'a, Coords>> {
-        debug!("SparseTableSlice::filled_at");
         self.shape().validate_axes(&axes)?;
 
-        if axes.is_empty() || self.is_empty(&txn).await? {
+        if axes.is_empty() {
             return Ok(Box::pin(stream::empty()));
         }
 
@@ -390,8 +413,8 @@ where
         let shape = axes.iter().map(|x| shape[*x]).collect::<Vec<u64>>();
         let source_axes = (0..self.source.ndim()).collect();
         let rebase = self.rebase;
-        let coords = filled_at::<FD, FS, D, T, _>(&txn, source_axes, self.table).await?;
-        let coords = CoordBlocks::new(coords, self.source.ndim(), PER_BLOCK)
+        let source_coords = filled_at::<FD, FS, D, T, _>(&txn, source_axes, self.table).await?;
+        let coords = CoordBlocks::new(source_coords, self.source.ndim(), PER_BLOCK)
             .map_ok(move |coords| rebase.map_coords(coords))
             .map_ok(move |coords| coords.get(&axes));
 
@@ -403,14 +426,17 @@ where
         self.table.count(*txn.id()).await
     }
 
-    async fn is_empty(&self, txn: &T) -> TCResult<bool> {
-        let mut rows = self.table.clone().rows(*txn.id()).await?;
-        rows.try_next().map_ok(|v| v.is_none()).await
-    }
-
     fn slice(self, bounds: Bounds) -> TCResult<Self::Slice> {
+        debug!("SparseTableSlice::slice {}", bounds);
         self.shape().validate_bounds(&bounds)?;
         let source_bounds = self.rebase.invert_bounds(bounds);
+
+        debug!(
+            "SparseTableSlice::slice from source {} with bounds {}",
+            self.source.shape(),
+            source_bounds
+        );
+
         self.source.slice(source_bounds)
     }
 
@@ -434,11 +460,18 @@ where
             self.shape().validate_coord(&coord)?;
             let dtype = self.dtype();
             let source_coord = self.rebase.invert_coord(&coord);
-            read_value_at(self.table, txn, source_coord, dtype).await
+            let (_, value) = read_value_at(self.table, txn, source_coord, dtype).await?;
+            Ok((coord, value))
         })
     }
 }
 
+impl<FD, FS, D, T> fmt::Display for SparseTableSlice<FD, FS, D, T> {
+    fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
+        f.write_str("a slice of a sparse Tensor's underlying Table")
+    }
+}
+
 struct SparseTableVisitor<FD, FS, D, T>
 where
     FD: File<Array>,
